/*GPL*START*
 * 
 * Copyright (C) 1998 by Johannes Overmann <overmann@iname.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * *GPL*END*/  

#include <qapp.h>
#include <qpainter.h>
#include <qdrawutl.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <qmsgbox.h> 
#include "qjocakediagram.h"
#include "mainwidget.h"
#include "tappconfig.h"


void MainWidget::want_quit() {
  if(!QMessageBox::query("about to quit", "All load-graph-data will be lost.\nAre you sure to quit?\n", "Oops! Don't quit!", "Quit"))
	 qApp->quit();
}


void MainWidget::popact(int id) {
  //   printf("popact(%d)\n", id);
   int i;
   switch(mode) {
    case M_GRAPH:
      if(id < grp.num()) {
	 for(i=0; i < m.num(); i++) 
	   if(m[i].group == grp[id]) m[i].graph = true;
      } else { // all
	 for(i=0; i < m.num(); i++) 
	   m[i].graph = true;
      }
		dmode=(MODE_T)-1;
      disp_graph();
      break;

	case M_LOAD:
	  //	  printf("command(%s)\n", *load_pop_str[id]);
	  if(!fork()) {
		 system(load_pop_str[id]);
		 exit(0);
	  } 
	  break;

    default:
      break;
   }
}


void MainWidget::mouseMoveEvent(QMouseEvent *e) {
  int i,j;
  
  if(mode==M_LOAD) {
	 for(i=0; i < load_rect.num(); i++) {
		if(load_rect[i].contains(e->pos())) break;
	 }
	 if(load_act[i](0,5)=="graph") {
		j = mustfindmach(load_act[i](6, TString::END));
		if((j>=0) && graph_drag) {
		  if(m[j].graph != graph_value) {
			 m[j].graph = graph_value;
			 //repaint(0,0,-1,-1,false);
			 draw_load(true);
		  }
		}
	 }	 
  }
}


void MainWidget::mouseReleaseEvent(QMouseEvent *) {
  if(mode==M_LOAD) {
	 graph_drag = false;
  }
}


void MainWidget::mousePressEvent(QMouseEvent *e) {
  if(pop) delete pop;
  pop=0;
  int i,j;
  switch(mode) {
  case M_GRAPH:
	 // setup popup
	 if(pop) delete pop;
	 if(pop1) delete pop1;
	 pop = new QPopupMenu;
	 for(i=0; i < grp.num(); i++) {
		pop->insertItem("Show group "  + grp[i]);
	 }
	 pop1 = new QPopupMenu;
	 for(i=0, j=0; i < m.num(); i++) {
		if(!m[i].graph) {
		  pop1->insertItem("Show " + m[i].name, i);
		  j++;
		}
	 }
	 if(j) {
		pop->insertSeparator();
		pop->insertItem("Show all");
		pop->insertItem("Show", pop1);
		connect(pop1, SIGNAL(activated(int)), SLOT(pop1act(int)));
	 }
	 connect(pop, SIGNAL(activated(int)), SLOT(popact(int)));
	 break;
	 
  case M_LOAD:
	  for(i=0; i < load_rect.num(); i++) {
		 if(load_rect[i].contains(e->pos())) break;
	  }
	  if(i<load_rect.num()) {
		 //		 printf("load_action(%s)\n", *load_act[i]);
		 if(load_act[i]=="sort arch") sort_arch();
		 if(load_act[i]=="sort name") sort_name();
		 if(load_act[i]=="sort group") sort_group();
		 if(load_act[i]=="sort ram") sort_ram();
		 if(load_act[i]=="sort load1") sort_load1();
		 if(load_act[i]=="sort load2") sort_load2();
		 if(load_act[i]=="sort load3") sort_load3();
		 if(load_act[i](0,5)=="graph") {
			j = mustfindmach(load_act[i](6, TString::END));
			if(j<0) j=0;
			m[j].graph = !m[j].graph;
			//			repaint(0,0,-1,-1,false);
			draw_load(true);
			graph_drag = true;
			graph_value = m[j].graph;
		 }
		 if(load_act[i](0,5)=="popup") {
			j = mustfindmach(load_act[i](6, TString::END));
			if(j>=0) {
			  FILE *f = fopen(tApp->getString("popup-file"), "r");
			  if(f==0) {
				 fprintf(stderr, "%s: error while opening file '%s' for reading!\n", *tApp->getString("application-name"), *tApp->getString("popup-file"));
			  } else {
				 // create popup
				 if(pop) delete pop;
				 pop = new QPopupMenu;
				 connect(pop, SIGNAL(activated(int)), SLOT(popact(int)));
				 load_pop_str.empty();

				 // add items
				 char buf[2000];
				 while(fgets(buf, sizeof(buf), f)) {
					if(!isspace(buf[0])) {
					  TString line(buf);
					  line.searchReplace("$host", m[j].name);
					  line.searchReplace("$group", m[j].group);
					  TArray<TString> a = split(line, ",", true, true);
					  if(a.num()>=2) {
						 load_pop_str += a[1];
						 pop->insertItem(a[0]);
					  }
					}
				 }
				 fclose(f);
			  }
			}			
		 }
	  }
	  break;

    default:
      break;
   }
   if(pop) pop->popup(mapToGlobal(e->pos()));
}


void MainWidget::paintEvent(QPaintEvent *) {
  if(mode==M_LOAD) {
	 disp_load();
  } else {
	 erase();
  }
}


static int get_str_dx(QPainter& p, const char *str) {
  return p.boundingRect(0,0,2048,128,AlignLeft,str).width();
}


static int get_str_dy(QPainter& p, const char *str) {
  return p.boundingRect(0,0,2048,128,AlignLeft,str).height();
}


void MainWidget::draw_load(bool only_graph) {
   if(dmode!=M_LOAD) {
      // clean up
      destroy_wid();
      dmode = mode = M_LOAD;      
      if(pop) delete pop;
      if(pop1) delete pop1;
		pop=pop1=0;
   }


	if(!only_graph) {
	// clean up
	//	erase();
	  load_rect.empty();
	  load_act.empty();
	}   

   // get painter
   QPainter p(this);
   p.setFont(QFont("helvetica", 10, QFont::Bold));      
   p.setPen(black);
   
   // get geom
   int min_dy = get_str_dy(p, "WgqiI/")-2;
   int numm = m.num();
   int max_yy = height()/min_dy;
   if(max_yy < 2) max_yy = 2;
   int xx = (numm+max_yy-2) / (max_yy-1);
   int yy = (numm+xx-1)/xx;
   int dx = width()/xx;
   int dy = height()/(yy+1);
   
   // get mach dimesions and load scale
   int max_name_dx = get_str_dx(p, "Host");
   int max_arch_dx = get_str_dx(p, "Arch");
   int max_ram_dx = get_str_dx(p, "Mem");
   int max_group_dx = get_str_dx(p, "Grp");
   int max_load_dx = get_str_dx(p, "15min");
   int max_graph_dx = dy;
   double max_load1=0;
   int i,j;
   for(i=0; i < m.num(); i++) {
      j = get_str_dx(p, m[i].name);
      if(j > max_name_dx) max_name_dx = j;
      j = get_str_dx(p, TString(m[i].ram));
      if(j > max_ram_dx) max_ram_dx = j;
      j = get_str_dx(p, m[i].group);
      if(j > max_group_dx) max_group_dx = j;
      j = get_str_dx(p, m[i].arch);
      if(j > max_arch_dx) max_arch_dx = j;
      if(m[i].load1 > max_load1) max_load1 = m[i].load1;
   }
   
   // draw
   int pad = 3;
   int name_cell_w = max_name_dx + 2*pad;
   int arch_cell_w = max_arch_dx + 2*pad;
   int ram_cell_w = max_ram_dx + 2*pad;
   int group_cell_w = max_group_dx + 2*pad;
   int load_cell_w = max_load_dx + 2*pad;
   int graph_cell_w = max_graph_dx;
   
   int x=0, y=0; // pos
   for(int cx=0; cx<xx; cx++, x+=dx, y=0) {
	  if(!only_graph) {
		 erase(x, y, dx, dy);
		 // line 
		 if(cx) {
			p.drawLine(x-1,dy,x-1,height());
		 }
		 
		 // head
		 
		 // host
		 QRect r(x, y, name_cell_w, dy);
		 p.setPen(black);
		 qDrawShadePanel(&p, r.x(), r.y(), r.width(), r.height(),
							  QWidget::colorGroup(), false, 2, NULL);
		 p.drawText(r, AlignHCenter|AlignVCenter, "Host");
		 load_rect += r;
		 load_act += "sort name";
		 r.moveBy(name_cell_w, 0);
		 
		 // arch
		 r.setSize(QSize(arch_cell_w, dy));
		 p.setPen(black);
		 qDrawShadePanel(&p, r.x(), r.y(), r.width(), r.height(),
							  QWidget::colorGroup(), false, 2, NULL);
		 p.drawText(r, AlignHCenter|AlignVCenter, "Arch");
		 load_rect += r;
		 load_act += "sort arch";
		 r.moveBy(arch_cell_w, 0);
		 
		 // ram
		 r.setSize(QSize(ram_cell_w, dy));
		 qDrawShadePanel(&p, r.x(), r.y(), r.width(), r.height(),
							  QWidget::colorGroup(), false, 2, NULL);
		 p.drawText(r, AlignHCenter|AlignVCenter, "Mem");
		 load_rect += r;
		 load_act += "sort ram";
		 r.moveBy(ram_cell_w, 0);
		 
		 // group
		 r.setSize(QSize(group_cell_w, dy));
		 qDrawShadePanel(&p, r.x(), r.y(), r.width(), r.height(),
							  QWidget::colorGroup(), false, 2, NULL);
		 p.drawText(r, AlignHCenter|AlignVCenter, "Grp");
		 load_rect += r;
		 load_act += "sort group";
		 r.moveBy(group_cell_w, 0);
		 
		 // graph
		 r.setSize(QSize(graph_cell_w, dy));
		 qDrawShadePanel(&p, r.x(), r.y(), r.width(), r.height(),
							  QWidget::colorGroup(), false, 2, NULL);
		 //      p.drawText(r, AlignHCenter|AlignVCenter, "G");
		 p.setBrush(white);
		 p.drawRect(r.x()+2, r.y()+2, r.width()-4, r.height()-4);
		 p.setPen(darkGreen);
		 p.drawLine(r.x()+3, r.y()+3, r.x()+2+4, r.y()+2+8);
		 p.drawLine(r.x()+2+4, r.y()+2+8, r.x()+2+7, r.y()+2+4);
		 p.drawLine(r.x()+2+7, r.y()+2+4, r.x()+dy-4, r.y()+dy-4);
		 p.setPen(black);
		 r.moveBy(graph_cell_w, 0);
		 
		 // load1
		 r.setSize(QSize(load_cell_w, dy));
		 qDrawShadePanel(&p, r.x(), r.y(), r.width(), r.height(),
							  QWidget::colorGroup(), false, 2, NULL);
		 p.drawText(r, AlignHCenter|AlignVCenter, "1min");
		 load_rect += r;
		 load_act += "sort load1";
		 r.moveBy(load_cell_w, 0);
		 
		 // load1
		 r.setSize(QSize(load_cell_w, dy));
		 qDrawShadePanel(&p, r.x(), r.y(), r.width(), r.height(),
							  QWidget::colorGroup(), false, 2, NULL);
		 p.drawText(r, AlignHCenter|AlignVCenter, "5min");
		 load_rect += r;
		 load_act += "sort load2";
		 r.moveBy(load_cell_w, 0);
		 
		 // load1
		 r.setSize(QSize(load_cell_w, dy));
		 qDrawShadePanel(&p, r.x(), r.y(), r.width(), r.height(),
							  QWidget::colorGroup(), false, 2, NULL);
		 p.drawText(r, AlignHCenter|AlignVCenter, "15min");
		 load_rect += r;
		 load_act += "sort load3";
		 r.moveBy(load_cell_w, 0);
		 
		 // bar
		 i = dx-3*load_cell_w-name_cell_w-group_cell_w-ram_cell_w-graph_cell_w-arch_cell_w;
		 if(i>=20) {
			r.setSize(QSize(i, dy));
			qDrawShadePanel(&p, r.x(), r.y(), r.width(), r.height(),
								 QWidget::colorGroup(), false, 2, NULL);
			p.drawText(r, AlignHCenter|AlignVCenter, "Load");
			load_rect += r;
			load_act += "sort load1";
			r.moveBy(i, 0);
		 }
	  }
	  
	  y+=dy;
	  
	  // machines
      for(int cy=0; cy<yy; cy++, y+=dy) {	
		  i = cx*yy + cy;
		  if(i >= m.num()) break;
		  if(!only_graph) {
			 erase(x, y, dx, dy);
			 char buf[100];
			 
			 // host
			 QRect r(x, y, name_cell_w, dy);
			 p.setPen(black);
			 r.moveBy(pad, 0);
			 p.drawText(r, AlignLeft|AlignVCenter, m[i].name);
			 r.moveBy(name_cell_w-pad, 0);
			 
			 // arch
			 r.setSize(QSize(arch_cell_w, dy));
			 if(m[i].arch!=UNK_ARCH) 
				p.drawText(r, AlignHCenter|AlignVCenter, m[i].arch);
			 else {
				p.setPen(red);
				p.drawText(r, AlignHCenter|AlignVCenter, "?");
				p.setPen(black);
			 }
			 r.moveBy(arch_cell_w, 0);
			 
			 // ram
			 r.setSize(QSize(ram_cell_w, dy));
			 if(m[i].ram>0) 
				p.drawText(r, AlignHCenter|AlignVCenter, TString(m[i].ram));
			 else {
				p.setPen(red);
				p.drawText(r, AlignHCenter|AlignVCenter, "n/a");
				p.setPen(black);
			 }
			 r.moveBy(ram_cell_w, 0);
			 
			 // group
			 r.setSize(QSize(group_cell_w, dy));
			 if(m[i].group!=UNK_GROUP) {
				p.drawText(r, AlignHCenter|AlignVCenter, m[i].group);
			 } else {
				p.setPen(red);
				p.drawText(r, AlignHCenter|AlignVCenter, "n/a");
				p.setPen(black);			 
			 }
			 r.moveBy(group_cell_w, 0);
			 
			 // graph
			 r.setSize(QSize(graph_cell_w, dy));
			 if(m[i].graph) {
				p.setBrush(white);
				p.drawRect(r.x()+2, r.y()+2, r.width()-4, r.height()-4);
				p.setPen(darkGreen);
				p.drawLine(r.x()+3, r.y()+3, r.x()+2+4, r.y()+2+8);
				p.drawLine(r.x()+2+4, r.y()+2+8, r.x()+2+7, r.y()+2+4);
				p.drawLine(r.x()+2+7, r.y()+2+4, r.x()+dy-4, r.y()+dy-4);
				p.setPen(black);
			 }
			 //		  p.drawText(r, AlignHCenter|AlignVCenter, m[i].graph?"X":".");
			 load_rect += r;
			 load_act += "graph "+m[i].name;
			 r.moveBy(graph_cell_w, 0);
			 
			 // load1
			 r.setSize(QSize(load_cell_w-pad, dy));
			 if(m[i].load1>=0.0) {
				sprintf(buf, "%4.2f", m[i].load1);
				p.drawText(r, AlignRight|AlignVCenter, buf);
			 } else {
				p.setPen(red);
				if(m[i].load1==UNK_LOAD) p.drawText(r, AlignRight|AlignVCenter, "n/a");	 
				p.setPen(black);
			 }
			 r.moveBy(load_cell_w, 0);
			 
			 // load2
			 r.setSize(QSize(load_cell_w-pad, dy));
			 if(m[i].load1>=0.0) {
				sprintf(buf, "%4.2f", m[i].load2);
				p.drawText(r, AlignRight|AlignVCenter, buf);
			 } else {
				p.setPen(red);
				if(m[i].load1==UNK_LOAD) p.drawText(r, AlignRight|AlignVCenter, "n/a");	 
				else p.drawText(r, AlignRight|AlignVCenter, "down");	 
				p.setPen(black);
			 }
			 r.moveBy(load_cell_w, 0);
			 
			 // load3
			 r.setSize(QSize(load_cell_w-pad, dy));
			 if(m[i].load1>=0.0) {
				sprintf(buf, "%4.2f", m[i].load3);
				p.drawText(r, AlignRight|AlignVCenter, buf);
			 } else {
				p.setPen(red);
				if(m[i].load1==UNK_LOAD) p.drawText(r, AlignRight|AlignVCenter, "n/a");	 
				p.setPen(black);
			 }
			 r.moveBy(load_cell_w, 0);
			 
			 // bar
			 j = dx-3*load_cell_w-name_cell_w-group_cell_w-ram_cell_w-graph_cell_w-arch_cell_w;
			 if(j>=20 && m[i].load1>0.0) {
				int coltab[]={0x00cc00, 0x11ffff, 0x0000ff, 0x4000c0, 0x800080, 
								  0xc00040, 0xff0000, 0xff4444, 0xff8888, 0xffcccc, 0xffffff};
				double bar = j-10;
				int col=0;
				double b;
				r.moveBy(5, 2);
				for(b=0.0; b < m[i].load1-1.0; b+=1.0, col++) {
				  p.fillRect(int(r.x()+bar*(b/max_load1)), r.y(), 
								 int(bar*1.0/max_load1)+1, dy-4, 
								 QBrush(QColor(coltab[col])));
				}
				p.fillRect(int(r.x()+bar*(b/max_load1)), r.y(), 
							  int(bar*(m[i].load1-b)/max_load1), dy-4, 
							  QBrush(QColor(coltab[col])));
				bar *= m[i].load1 / max_load1;
				r.setSize(QSize(int(bar), dy-4));
				p.setBrush(NoBrush);
				p.drawRect(r);
			 }		  
			 r.setRect(x,y,dx,dy);
			 load_rect += r;
			 load_act += "popup " + m[i].name;

		  } else { // only-graph

			 // calc pos
			 QRect r(x, y, name_cell_w, dy);
			 r.moveBy(name_cell_w + group_cell_w+
						 ram_cell_w+arch_cell_w, 0);

			 // graph
			 r.setSize(QSize(graph_cell_w, dy));
			 if(m[i].graph) {
				p.setBrush(white);
				p.drawRect(r.x()+2, r.y()+2, r.width()-4, r.height()-4);
				p.setPen(darkGreen);
				p.drawLine(r.x()+3, r.y()+3, r.x()+2+4, r.y()+2+8);
				p.drawLine(r.x()+2+4, r.y()+2+8, r.x()+2+7, r.y()+2+4);
				p.drawLine(r.x()+2+7, r.y()+2+4, r.x()+dy-4, r.y()+dy-4);
				p.setPen(black);
			 } else erase(r);
			 
		  }
      }
   }
}


void MainWidget::disp_load() {
  draw_load();
}


void MainWidget::disp_groups() {
   static QjoCakeDiagram *av;
   static QjoCakeDiagram *lo;
   
   if(dmode!=M_GROUPS) {
      // clean up
      destroy_wid();
      mode = M_GROUPS;   
		erase();
      av = new QjoCakeDiagram(this);
      lo = new QjoCakeDiagram(this);
      av->setTitle("Average Load per Group");
      lo->setTitle("Total Load per Group");
      lo->setFormat("<name>\n<value>", "%.1f", "%.2f");
      av->setFormat("<name>\n<value>", "%.1f", "%.2f");
      wid += av;
      wid += lo;
      lay = new QBoxLayout(this, QBoxLayout::LeftToRight, 5);
      lay->addWidget(av, 1);
      lay->addWidget(lo, 1);   
      connect(av, SIGNAL(pressed(const char *, const QPoint&)), SLOT(group_pop(const char *, const QPoint&)));
      connect(lo, SIGNAL(pressed(const char *, const QPoint&)), SLOT(group_pop(const char *, const QPoint&)));
   } else {
      av->clearSlices();
      lo->clearSlices();
   }
   
   int i;
   for(i=0; i < grp.num(); i++) {
      av->setSlice(grp[i], grphistl(grp[i])/grphist(grp[i]));
      lo->setSlice(grp[i], grphistl(grp[i]));
   }
   
   if(dmode!=M_GROUPS) {
      // show
      no_resize = true;
      lay->activate();
      no_resize = false;
      for(i=0; i < wid.num(); i++) wid[i]->show();
      dmode=M_GROUPS;
   } else {
      av->repaint(0,0,-1,-1,false);
      lo->repaint(0,0,-1,-1,false);
   }
}


